
import jinja2
from typing import Dict

init_sql_schema_script = """
BEGIN;
CREATE TABLE IF NOT EXISTS {{ tablename }}(
    {%- for name, field in sql_fields.items() %}
    {{ name }} {{ field }}{% if not loop.last %},{% endif %}
    {%- endfor %}
);
CREATE VIRTUAL TABLE IF NOT EXISTS {{ tablename }}_fts
USING FTS5(
    {%- for name in sql_fields %}
    {{ name }},
    {%- endfor %}
    tokenize = "{{ tokenize or 'trigram' }}",
    content = "{{ tablename }}"
);
CREATE TRIGGER IF NOT EXISTS {{ tablename }}_ai AFTER INSERT ON {{ tablename }} BEGIN
    INSERT INTO {{ tablename }}_fts({% for name in sql_fields %}{{ name }}{% if not loop.last %}, {% endif %}{% endfor %})
    VALUES({% for name in sql_fields %}new.{{ name }}{% if not loop.last %}, {% endif %}{% endfor %});
END;
CREATE TRIGGER IF NOT EXISTS {{ tablename }}_ad AFTER DELETE ON {{ tablename }} BEGIN
    INSERT INTO {{ tablename }}_fts({{ tablename }}_fts, {% for name in sql_fields %}{{ name }}{% if not loop.last %}, {% endif %}{% endfor %})
    VALUES('delete', {% for name in sql_fields %}old.{{ name }}{% if not loop.last %}, {% endif %}{% endfor %});
END;
COMMIT;
"""


index_sql_query = """
INSERT INTO {{ tablename }}({{ sql_insert }}) VALUES ({{ sql_insert_q }})
ON CONFLICT({{ sql_pkey }}) DO UPDATE 
SET
    {%- for name, field in sql_fields.items() %}
    {%- if name != sql_pkey %}
    {{ name }} = excluded.{{ name }}{% if not loop.last %},{% endif %}
    {%- endif %}
    {%- endfor %}
"""

insert_sql_query = """
INSERT INTO {{ tablename }}({{ sql_insert }}) VALUES ({{ sql_insert_q }})
ON CONFLICT({{ sql_pkey }}) DO UPDATE
SET
    {%- for name in sql_insert_fields %}
    {%- if name != sql_pkey and name in sql_insert_fields %}
    {{ name }} = excluded.{{ name }}{% if not loop.last %},{% endif %}
    {%- endif %}
    {%- endfor %}
"""


search_sql_query = """
SELECT {% if return_fields %}{% for field in return_fields %}{{ field }}{% if not loop.last %}, {% endif %}{% endfor %}{% elif return_id_only %}{{ sql_pkey }}{% else %}*{% endif %}, rank
FROM {{ tablename }}_fts 
WHERE {{ query }}
ORDER BY rank
{%- if limit is defined and limit %}
LIMIT {{ limit }}
{%- endif %}
{%- if skip is defined and skip %}
OFFSET {{ skip }}
{%- endif %};
"""

# Refresh the underlying item
refresh_sql_query = """
SELECT * FROM {{ tablename }} WHERE {{ sql_pkey }} = ?
"""

count_total_sql_query = """
SELECT COUNT(*) FROM {{ tablename }} WHERE {{ sql_pkey }} is not null
"""

update_field_attribute_sql_query = """
UPDATE {{ tablename }} 
SET
    {{ name }} = ?
WHERE {{ sql_pkey }} = ?
"""
"""
SELECT rowid FROM {{ tablename }}_fts WHERE {{ sql_pkey }} = ?
    DELETE FROM {{ tablename }}_fts WHERE rowid = rowid
"""

# Delete doesn't currently work.
delete_row_sql_query = """
DELETE FROM {{ tablename }} WHERE {{ sql_pkey }} = ?
"""

SqliteTemplates: Dict[str, jinja2.Template] = {
    'init': jinja2.Template(init_sql_schema_script),
    'index': jinja2.Template(index_sql_query),
    'insert': jinja2.Template(insert_sql_query),
    'search': jinja2.Template(search_sql_query),
    'refresh': jinja2.Template(refresh_sql_query),
    'count_total': jinja2.Template(count_total_sql_query),
    'update_field_attribute': jinja2.Template(update_field_attribute_sql_query),
    'delete': jinja2.Template(delete_row_sql_query),
}


import re
from typing import Optional

alphanumeric_pattern = re.compile(r'[^a-zA-Z0-9]')

def normalize_sql_text(text: Optional[str]) -> Optional[str]:
    """
    Normalizes the text to be indexed/queried
    """
    if text is None or not text.strip(): return None
    return alphanumeric_pattern.sub(' ', text).replace('  ', ' ').replace('  ', ' ').strip()



"""
Authentication Objects
"""
import base64
import httpx
from lazyops.libs import lazyload
from typing import Optional, List, Dict, Any, Union, Iterable, Generator, AsyncGenerator
from .base import BaseModel
from .user_data import AZUserData
from .claims import APIKeyJWTClaims


if lazyload.TYPE_CHECKING:
    # import niquests
    from starlette.datastructures import Headers
    from starlette.requests import Request
    from niquests.models import Headers, Request as HTTPRequest
    from ..configs import AuthZeroSettings
    from ..flows.tokens import ClientCredentialsFlow, APIClientCredentialsFlow, AccessToken
# else:
    # niquests = lazyload.LazyLoad("niquests")

class AuthZeroTokenAuth(httpx.Auth):
    """
    [niquests/httpx] Implements the AuthBase for the AuthZero HTTP Token

    - It will always use the `Authorization` header and automatically handle token expiration
    """

    def __init__(
        self,
        token_flow: Union['ClientCredentialsFlow', 'APIClientCredentialsFlow'],
        x_api_key: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ):
        self.token_flow = token_flow
        self.x_api_key = x_api_key
        self.headers = headers or {}

    """
    Implements the niquests logic
    """
    def __eq__(self, other: Union[str, 'AccessToken', 'AuthZeroTokenAuth']) -> bool:
        """
        Compares the token to the other
        """
        if isinstance(other, str): return self.token_flow.token == other
        if hasattr(other, 'token_flow'): return self.token_flow.token == other.token_flow.token
        return self.token_flow.token == getattr(other, "token", None)
    
    def __ne__(self, other: Union[str, 'AccessToken', 'AuthZeroTokenAuth']) -> bool:
        """
        Compares the token to the other
        """
        return not self == other

    def __call__(self, r: 'HTTPRequest') -> 'HTTPRequest':
        """
        Implements the AuthZero HTTP Credential Injection
        """
        r.headers['Authorization'] = f'Bearer {self.token_flow.token}'
        if self.x_api_key is not None: r.headers['x-api-key'] = self.x_api_key
        if self.headers: 
            for key, value in self.headers.items():
                if key not in r.headers: r.headers[key] = value
        # print(r.headers)
        return r
    
    """
    Implements the httpx logic
    """

    def auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        """
        Implements the httpx auth flow
        """
        request.headers['Authorization'] = f'Bearer {self.token_flow.token}'
        if self.x_api_key is not None: request.headers['x-api-key'] = self.x_api_key
        if self.headers: 
            for key, value in self.headers.items():
                if key not in request.headers: request.headers[key] = value
        yield request



class AuthObject(BaseModel):
    auth_token: Optional[str] = None
    x_api_key: Optional[str] = None

    model_config = {'arbitrary_types_allowed': True}
    
    @property
    def has_auth_token(self) -> bool:
        """
        Returns True if the Auth Token is Present
        """
        return bool(self.auth_token)
    
    @property
    def has_x_api_key(self) -> bool:
        """
        Returns True if the X-API-Key is Present
        """
        return bool(self.x_api_key)
    
    @property
    def has_both_auth(self) -> bool:
        """
        Returns True if the Auth Token and X-API-Key are Present
        """
        return self.has_auth_token and self.has_x_api_key
    
    @classmethod
    def get_auth_token(cls, data: Union['Headers', Dict[str, str]], settings: Optional['AuthZeroSettings'] = None) -> Optional[str]:
        """
        Returns the Auth Token from the Headers or Cookies
        """
        settings = settings or cls.get_settings()
        authorization_header_value = data.get(settings.authorization_header)
        if authorization_header_value:
            scheme, _, param = authorization_header_value.partition(" ")
            if scheme.lower() == settings.authorization_scheme and not param.startswith("apikey:"):
                return param
    
    @classmethod
    def get_x_api_key(cls, data: Union['Headers', Dict[str, str]], settings: Optional['AuthZeroSettings'] = None) -> Optional[str]:
        """
        Returns the API Key from the Headers or Cookies
        """
        # We're supporting `Authorization: Basic apikey:secret` for SPARQL Server Implementations
        settings = settings or cls.get_settings()
        if api_key := data.get(settings.apikey_header):
            return api_key
        authorization_header_value = data.get(settings.authorization_header)
        if authorization_header_value:
            scheme, _, param = authorization_header_value.partition(" ")
            if scheme.lower() in {"basic", "bearer"}:
                if scheme.lower() == 'basic':
                    param = base64.b64decode(param).decode("utf-8")
                if "apikey:" not in param:
                    return None
                _, api_key = param.split(":", 1)
                return api_key


    @classmethod
    def from_headers(cls, headers: 'Headers', settings: Optional['AuthZeroSettings'] = None) -> 'AuthObject':
        """
        Returns an AuthObject from the Headers
        """
        settings = settings or cls.get_settings()
        return cls(
            auth_token = cls.get_auth_token(data = headers, settings = settings),
            x_api_key = cls.get_x_api_key(data = headers, settings = settings),
        )


    @classmethod
    def extract_auth_token(cls, *data: Iterable[Union['Headers', Dict[str, str]]], settings: Optional['AuthZeroSettings'] = None) -> Optional[str]:
        """
        Extract the Auth Token from the Headers, Cookies, or Dict
        """
        settings = settings or cls.get_settings()
        for item in data:
            if token := cls.get_auth_token(data = item, settings = settings): return token
        return None

    @classmethod
    def extract_x_api_key(cls, *data: Iterable[Union['Headers', Dict[str, str]]], settings: Optional['AuthZeroSettings'] = None) -> Optional[str]:
        """
        Extract the API Key from the Headers, Cookies, or Dict
        """
        settings = settings or cls.get_settings()
        for item in data:
            if api_key := cls.get_x_api_key(data = item, settings = settings): return api_key
        return None

    @classmethod
    def from_request(cls, request: 'Request', settings: Optional['AuthZeroSettings'] = None) -> 'AuthObject':
        """
        Returns an AuthObject from the Request
        """
        settings = settings or cls.get_settings()
        return cls(
            auth_token = cls.extract_auth_token(request.headers, request.cookies, settings = settings),
            x_api_key = cls.extract_x_api_key(request.headers, request.cookies, settings = settings),
        )

    @classmethod
    def from_items(cls, *items: Iterable[Dict[str, str]], settings: Optional['AuthZeroSettings'] = None) -> 'AuthObject':
        """
        Returns an AuthObject from the Iterable of Dicts
        """
        settings = settings or cls.get_settings()
        return cls(
            auth_token = cls.extract_auth_token(*items, settings = settings),
            x_api_key = cls.extract_x_api_key(*items, settings = settings),
        )
    


class APIKeyData(BaseModel):
    """
    The stored API Key Data
    """
    user_data: AZUserData
    claims: APIKeyJWTClaims



"""
Common Types
"""

from enum import Enum
from typing import Optional, List, Dict, Any, Union, Literal, TYPE_CHECKING

ValidationMethod = Literal['api_key', 'auth_token']
UserType = Literal['user', 'api_client', 'service', 'system', 'admin', 'allowed']



from __future__ import annotations


"""
V2 Implementation of AioSQLite
"""


__all__ = ["connect", "Connection", "Cursor"]

import abc
import asyncio
import sqlite3
import contextlib
import functools
from concurrent.futures import ThreadPoolExecutor
from collections.abc import Callable, Sequence
from functools import partial, wraps
from typing import Any, Literal, Optional, overload, Awaitable, TypeVar
from anyio import CapacityLimiter, to_thread


T = TypeVar('T')


class PoolObject(abc.ABC):
    """
    Base Pool Object
    """

    def __init__(
        self, 
        *args,
        max_workers: Optional[int] = None,
        pool: Optional[ThreadPoolExecutor] = None,
        concurrent: Optional[bool] = False,
        **kwargs,
    ) -> None:
        if pool is None: pool = ThreadPoolExecutor(max_workers = max_workers)
        self.pool = pool
        self.concurrent = concurrent
    
    def run_async(
        self, 
        func: Callable[..., T],
        *args, 
        **kwargs
    ) -> Awaitable[T]:
        """
        Wraps the function in a thread pool
        """
        blocking = functools.partial(func, *args, **kwargs)
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(self.pool, blocking)
        # future = self.pool.submit(func, *args, **kwargs)
        # return asyncio.wrap_future(future)

    
    async def run_async(
        self, 
        func: Callable[..., T],
        *args, 
        **kwargs
    ) -> Awaitable[T]:
        """
        Wraps the function in a thread pool
        """
        future = self.pool.submit(func, *args, **kwargs)
        f = asyncio.wrap_future(future)
        await f
        return f.result()
        
    

class Connection(PoolObject):
    def __init__(
        self, 
        _real_connection: sqlite3.Connection, 
        max_workers: Optional[int] = None,
        concurrent: Optional[bool] = False,
        **kwargs,
    ) -> None:
        
        super().__init__(max_workers = max_workers, concurrent = concurrent, **kwargs)
        self._real_connection = _real_connection
    

    @wraps(sqlite3.Connection.close)
    async def close(self):
        """
        Closes the connection
        """
        return await self.run_async(self._real_connection.close)

    @wraps(sqlite3.Connection.commit)
    async def commit(self):
        """
        Commits the connection
        """
        return await self.run_async(self._real_connection.commit)

    @wraps(sqlite3.Connection.rollback)
    async def rollback(self):
        """
        Rolls back the connection
        """
        return await self.run_async(self._real_connection.rollback)

    async def cursor(self, factory: Callable[[sqlite3.Connection], sqlite3.Cursor] = sqlite3.Cursor) -> Cursor:
        """
        Creates a cursor
        """
        real_cursor = await self.run_async(self._real_connection.cursor, factory)
        return Cursor(real_cursor, pool = self.pool)

    async def __aenter__(self) -> "Connection":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()


class Cursor(PoolObject):
    def __init__(
        self, 
        real_cursor: sqlite3.Cursor, 
        pool: Optional[ThreadPoolExecutor] = None,
        **kwargs,
    ) -> None:
        self._real_cursor = real_cursor
        super().__init__(pool = pool, **kwargs)


    @property
    def description(self) -> Any:
        """
        Returns the description
        """
        return self._real_cursor.description

    @property
    def rowcount(self) -> int:
        """
        Returns the row count
        """
        return self._real_cursor.rowcount

    @property
    def arraysize(self) -> int:
        """
        Returns the array size
        """
        return self._real_cursor.arraysize
    
    @property
    def lastrowid(self) -> int:
        """
        Returns the last row id
        """
        return self._real_cursor.lastrowid

    @wraps(sqlite3.Cursor.close)
    async def close(self) -> None:
        """
        Closes the cursor
        """
        await self.run_async(self._real_cursor.close)

    @wraps(sqlite3.Cursor.execute)
    async def execute(self, sql: str, parameters: Sequence[Any] = (), /) -> Cursor:
        """
        Executes the sql
        """
        real_cursor = await self.run_async(self._real_cursor.execute, sql, parameters)
        return Cursor(real_cursor, pool = self.pool)

    @wraps(sqlite3.Cursor.executemany)
    async def executemany(self, sql: str, parameters: Sequence[Any], /) -> Cursor:
        """
        Executes the sql
        """
        real_cursor = await self.run_async(self._real_cursor.executemany, sql, parameters)
        return Cursor(real_cursor, pool = self.pool)


    @wraps(sqlite3.Cursor.executescript)
    async def executescript(self, sql_script: str, /) -> Cursor:
        """
        Executes the sql script
        """
        real_cursor = await self.run_async(self._real_cursor.executescript, sql_script)
        return Cursor(real_cursor, pool = self.pool)

    @wraps(sqlite3.Cursor.fetchone)
    async def fetchone(self) -> tuple[Any, ...] | None:
        """
        Fetches the first row
        """
        return await self.run_async(self._real_cursor.fetchone)

    @wraps(sqlite3.Cursor.fetchmany)
    async def fetchmany(self, size: int) -> list[tuple[Any, ...]]:
        """
        Fetches the rows
        """
        return await self.run_async(self._real_cursor.fetchmany, size)

    @wraps(sqlite3.Cursor.fetchall)
    async def fetchall(self) -> list[tuple[Any, ...]]:
        """
        Fetches the rows
        """
        return await self.run_async(self._real_cursor.fetchall)
    
    async def __aenter__(self) -> "Cursor":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

@overload
async def connect(
    database: str,
    timeout: Optional[float] = None,
    detect_types: Optional[int] = None,
    isolation_level: Literal['DEFERRED', 'IMMEDIATE', 'EXCLUSIVE'] = None,
    cached_statements: Optional[int] = 128,
    uri: Optional[bool] = False,
    limit: Optional[int] = None,
) -> Connection:
    """
    Connects to the database
    """
    ...


async def connect(
    database: str,
    max_workers: Optional[int] = None,
    concurrent: Optional[bool] = False,
    **kwargs,
) -> Connection:
    """
    Connects to the database
    """
    _ = kwargs.pop('check_same_thread', None)
    real_connection = await to_thread.run_sync(
        partial(sqlite3.connect, database, check_same_thread=False, **kwargs)
    )
    return Connection(real_connection, max_workers = max_workers, concurrent = concurrent)


"""
SQLite Utilities and Mixins
"""

from .mixins import SQLiteModelMixin
from .index import SQLiteIndex
from .utils import normalize_sql_text

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .mixins import SQLResultT





_base_worker_index: int = None

def get_base_worker_index() -> int:
    """
    Gets the base worker index
    """
    global _base_worker_index
    if _base_worker_index is None:
        from lazyops.utils.system import is_in_kubernetes, get_host_name
        if is_in_kubernetes() and get_host_name()[-1].isdigit():
            _base_worker_index = int(get_host_name()[-1])
        else:
            _base_worker_index = 0
    return _base_worker_index



from __future__ import annotations

"""
The Base Configuration Schema
"""

import os
import contextlib
from enum import Enum
from pathlib import Path

from pydantic import model_validator
from lazyops.utils.logs import Logger, null_logger
from lazyops.imports._pydantic import BaseAppSettings, BaseModel
# from lazyops.libs.fastapi_utils import GlobalContext
from lazyops.libs.abcs.state import GlobalContext
from lazyops.libs.fastapi_utils.types.persistence import TemporaryData
from typing import List, Optional, Dict, Any, Callable, Union, Type, TYPE_CHECKING
from .types import AppEnv, get_app_env

if TYPE_CHECKING:
    from .ctx import ApplicationContext
    from lazyops.utils.logs import Logger


class AppSettings(BaseAppSettings):
    """
    App Settings
    """

    ingress_base: Optional[str] = None
    ingress_domain: Optional[str] = None
    debug_enabled: Optional[bool] = False
    app_env: Optional[AppEnv] = None

    if TYPE_CHECKING:
        ctx: ApplicationContext
    else:
        ctx: Optional[Any] = None
    
    @model_validator(mode = 'after')
    def validate_app_settings(self):
        """
        Validates the app environment
        """
        from .ctx import ApplicationContextManager
        self.ctx = ApplicationContextManager.get_ctx(
            self.module_name,
            ingress_domain = self.ingress_domain,
            ingress_base = self.ingress_base,
        )
        self.app_env = self.ctx.app_env
        if self.__class__.__name__.lower() == f'{self.__module__}settings':
            self.ctx.register_settings(self)
            from .lazy import register_module_settings
            register_module_settings(self.__module__, self)
        return self
    
    @property
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return self.ctx.logger
    
    @property
    def null_logger(self) -> 'Logger':
        """
        Returns the null logger
        """
        return null_logger
    
    @property
    def autologger(self) -> 'Logger':
        """
        Returns the auto logger
        """
        return self.logger if (self.is_development_env or self.debug_enabled) else self.null_logger
    
    @property
    def temp_data(self) -> 'TemporaryData':
        """
        Retrieves the temporary data
        """
        return self.ctx.temp_data
    
    @property
    def config_path(self) -> Path:
        """
        Retrieves the config path
        """
        return self.ctx.config_path

    @property
    def is_leader_process(self) -> bool:
        """
        Returns if this is the leader process
        """
        return self.ctx.is_leader_process
    

    @property
    def is_primary_server_process(self) -> bool:
        """
        Returns if this is the primary server process
        """
        return self.ctx.is_primary_server_process
    
    @property
    def is_primary_node(self) -> bool:
        """
        Returns whether or not this is the primary node
        """
        return self.host_name[-1] == '0' if self.in_k8s else True

    @property
    def is_local_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL] and not self.in_k8s
    
    @property
    def is_production_env(self) -> bool:
        """
        Returns whether the environment is production
        """
        return self.app_env == AppEnv.PRODUCTION and self.in_k8s

    @property
    def is_development_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL, AppEnv.CICD]
    
    def set_app_env(self, env: AppEnv) -> None:
        """
        Sets the app environment
        """
        self.app_env = self.app_env.from_env(env)

    def get_assets(
        self, 
        *path_parts,
        load_file: Optional[bool] = False,
        loader: Optional[Callable] = None,
        **kwargs
    ) -> Union[Path, Any, List[Path], List[Any], Dict[str, Path], Dict[str, Any]]:
        """
        Retrieves the assets

        args:
            path_parts: path parts to the assets directory (default: [])
            load_file: load the file (default: False)
            loader: loader function to use (default: None)
            **kwargs: additional arguments to pass to `get_module_assets`
        """
        return self.ctx.get_assets(*path_parts, load_file = load_file, loader = loader, **kwargs)
    
    def import_assets(
        self, 
        *path_parts,
        model: Optional[Type['BaseModel']] = None,
        load_file: Optional[bool] = False,
        **kwargs
    ) -> Union[Dict[str, Any], List[Any]]:
        """
        Import assets from a module.

        args:
            path_parts: path parts to the assets directory (default: [])
            model: model to parse the assets with (default: None)
            load_file: load the file (default: False)
            **kwargs: additional arguments to pass to import_module_assets
        """
        return self.ctx.import_assets(*path_parts, model = model, load_file = load_file, **kwargs)
    

    def get_app_env_file(
        self,
        name: Optional[str] = None, 
        required: Optional[bool] = False, 
        allow_default: Optional[bool] = True,
        configs_path: Optional[Path] = None,
    ) -> Optional[Path]:
        """
        Retrieves the app environment file

        Only valid for local/dev environments
        """
        return self.ctx.get_app_env_file(name = name, required = required, allow_default = allow_default, configs_path = configs_path)
    

    def get_app_default_file(
        self,
        name: Optional[str] = None, 
        required: Optional[bool] = False, 
        suffix: Optional[str] = None,
        configs_path: Optional[Path] = None,
    ) -> Optional[Path]:
        """
        Retrieves the app environment file

        Only valid for local/dev environments
        """
        return self.ctx.get_app_default_file(name = name, required = required, suffix = suffix, configs_path = configs_path)

        
    def get_app_ingress(
        self,
        app_host: Optional[str] = None,
        app_port: Optional[int] = None,
    ) -> str:
        """
        Retrieves the app ingress
        """
        return self.ctx.get_app_ingress(app_host = app_host, app_port = app_port)

    @property
    def global_ctx(self) -> GlobalContext:
        """
        Returns the global context
        """
        return GlobalContext
    
    def register_on_close(self, func: Callable, *args, **kwargs):
        """
        Registers a function to be called on close
        """
        self.global_ctx.register_on_close(func, *args, **kwargs)

    




"""
Configuration Types
"""

import os
from enum import Enum
from typing import Any

class AppEnv(str, Enum):
    CICD = "cicd"
    DEVELOPMENT = "development"
    STAGING = "staging"
    PRODUCTION = "production"
    LOCAL = "local"

    @classmethod
    def from_env(cls, env_value: str) -> "AppEnv":
        """
        Get the app environment from the environment variables
        """
        env_value = env_value.lower()
        if "cicd" in env_value or "ci/cd" in env_value: return cls.CICD
        if "prod" in env_value: return cls.PRODUCTION
        if "dev" in env_value: return cls.DEVELOPMENT
        if "staging" in env_value: return cls.STAGING
        if "local" in env_value: return cls.LOCAL
        raise ValueError(f"Invalid app environment: {env_value} ({type(env_value)})")
    
    @classmethod
    def from_hostname(cls, hostname: str) -> "AppEnv":
        """
        Get the app environment from the hostname
        """
        hostname = hostname.lower()
        if "dev" in hostname: return cls.DEVELOPMENT
        if "staging" in hostname: return cls.STAGING
        return cls.LOCAL if "local" in hostname else cls.PRODUCTION
        
    
    @classmethod
    def from_module_name(cls, module_name: str) -> 'AppEnv':
        """
        Retrieves the app environment
        """
        module_name = module_name.replace(".", "_").upper()
        for key in {
            "SERVER_ENV",
            f"{module_name}_ENV",
            "APP_ENV",
            "ENVIRONMENT",
        }:
            if env_value := os.getenv(key):
                return cls.from_env(env_value)

        from lazyops.utils.system import is_in_kubernetes, get_host_name
        if is_in_kubernetes():
            hn = get_host_name()
            try:
                parts = hn.split("-")
                return cls.from_env(parts[2]) if len(parts) > 3 else cls.PRODUCTION
            except Exception as e:
                return cls.from_hostname(hn)
        
        return cls.LOCAL
    
    def __eq__(self, other: Any) -> bool:
        """
        Equality operator
        """
        if isinstance(other, str): return self.value == other.lower()
        return self.value == other.value if isinstance(other, AppEnv) else False

    @property
    def is_devel(self) -> bool:
        """
        Returns True if the app environment is development
        """
        return self in [self.LOCAL, self.CICD, self.DEVELOPMENT]

    @property
    def is_local(self) -> bool:
        """
        Returns True if the app environment is local
        """
        return self in [self.LOCAL, self.CICD]

    @property
    def name(self) -> str:
        """
        Returns the name in lower
        """
        return self.value.lower()


    


def get_app_env(
    module_name: str,
) -> AppEnv:
    """
    Retrieves the app environment
    """
    module_name = module_name.replace(".", "_").upper()
    for key in {
        "SERVER_ENV",
        f"{module_name}_ENV",
        "APP_ENV",
        "ENVIRONMENT",
    }:
        if env_value := os.getenv(key):
            return AppEnv.from_env(env_value)

    from lazyops.utils.system import is_in_kubernetes, get_host_name
    if is_in_kubernetes():
        # Name should be
        # scout-<service>-<index>
        # or 
        # scout-<service>-<env>-<index>
        hn = get_host_name()
        try:
            parts = hn.split("-")
            return AppEnv.from_env(parts[1]) if len(parts) > 2 else AppEnv.PRODUCTION
        except Exception as e:
            return AppEnv.from_hostname(hn)
        # parts = get_host_name().split("-")
        # return AppEnv.from_env(parts[2]) if len(parts) > 3 else AppEnv.PRODUCTION
    
    return AppEnv.LOCAL